

from typing import Callable, Iterable, Tuple, Union, Dict, Optional, Any

import torch, abc
import numpy as np
from functools import partial, lru_cache, reduce

from .utils import safe_cast
from .transformation import Identifier, Indexer, StandardScaler, MinMaxScaler
from .tags import FieldTags, SPARSE, DENSE
from ..utils import warnLogger


__all__ = ['Field', 'BufferField', 'SparseField', 'DenseField', 'FieldTuple', 'FieldList', 'FieldModule']


TRANSFORMATIONS = {
    "none": Identifier,
    'label2index': Indexer,
    'minmax': MinMaxScaler,
    'standard': StandardScaler,
}


class Field(metaclass=abc.ABCMeta):
    r"""
    Fielding data by column.

    Parametesr:
    -----------
    name: str, optional 
        Name of this field.
    tags: set 
        A set of FieldTags.
    dtype: Any 
        Data type of the field.
    caster: Callable 
        Function to convert elements to specified dtype (int, float or str) with na_value.
    na_value: Union[None, str, int, float] 
        The fill value for null/missing values.

    Examples:
    ---------
    >>> from freerec.data.tags import USER, ITEM, ID, TARGET
    >>> User = SparseField('User', None, int, tags=(USER, ID))
    >>> Item = SparseField('Item', None, int, tags=(ITEM, ID))
    >>> Target = SparseField('Label', 0, int, transformer='none', tags=TARGET)
    """

    def __init__(
        self, data: Any = None, 
        tags: Union[FieldTags, Iterable[FieldTags]] = tuple()
    ) -> None:
        self.__tags = set()
        if isinstance(tags, FieldTags):
            self.add_tag(tags)
        else:
            self.add_tag(*tags)
        self.data = data

    def add_tag(self, *tags: FieldTags) -> None:
        r"""
        Add some tags.

        Parameters:
        -----------
        *tags: FieldTags 
            The tags to add.

        Examples:
        ---------
        >>> from freerec.data.tags import ID
        >>> User = SparseField('User', -1, int)
        >>> User.add_tag(ID)
        """
        for tag in tags:
            if not isinstance(tag, FieldTags):
                warnLogger(f"FieldTags is expected but {type(tags)} received ...")
            self.__tags.add(tag)

    @property
    def tags(self) -> set:
        return self.__tags

    def match(self, *tags: FieldTags):
        r"""
        If current field matches the given tags, return True.

        Parametesr:
        -----------
        *tags (FieldTags): The tags to match.

        Returns:
        --------
        bool: True if all given tags are matched, False otherwise.

        Examples:
        ---------
        >>> from freerec.data.tags import ID
        >>> User = SparseField('User', -1, int)
        >>> User.add_tag(ID)
        >>> User.match(ID)
        True
        >>> User.match(ID, Feature)
        False
        >>> User.match(Feature)
        False
        """
        return all(tag in self.__tags for tag in tags)

    def buffer(
        self, data: Any = None,
        tags: Union[FieldTags, Iterable[FieldTags]] = tuple()
    ) -> 'BufferField':
        r"""
        Return a new BufferField with the given or inherited data.

        Parameters:
        -----------
        data: Any
            Any column data.
        tags: Union[FieldTags, Iterable[FieldTags]] 
            Tags for filtering.

        Notes:
        ------
        The tags will also be inherited.
        """
        buffer_ = BufferField(data, tags, root=self)
        return buffer_


class BufferField(Field):
    r"""
    For buffering data, which should be re-created once the data changes.
    
    Parameters:
    -----------
    data: Any
        Any column data.
    tags: Union[FieldTags, Iterable[FieldTags]] 
        Tags for filtering.
    root: Field, optional
        Inherit some attribuites from the root.
    """

    def __init__(
        self, data: Any, 
        tags: Union[FieldTags, Iterable[FieldTags]] = tuple(),
        *, root: Optional[Field] = None
    ) -> None:
        super().__init__(data, tags)
        self.data = data
        if isinstance(tags, FieldTags):
            self.add_tag(tags)
        else:
            self.add_tag(*tags)
        self.inherit(root)

    def inherit(self, root: Union[None, 'BufferField', 'FieldModule']):
        if root is None:
            pass
        elif root.match(SPARSE):
            self.count = root.count
            self.add_tag(*root.tags)
        elif root.match(DENSE):
            self.add_tag(*root.tags)
        else:
            raise ValueError(
                f"root should be `None|BufferField|FieldMoudle' but {type(root)} received ..."
            )

    def __getitem__(self, *args, **kwargs):
        return self.data.__getitem__(*args, **kwargs)

    def __iter__(self):
        yield from iter(self.data)

    def to(
        self, device: Optional[Union[int, torch.device]] = None,
        dtype: Optional[Union[torch.dtype, str]] = None,
        non_blocking: bool = False
    ):
        if isinstance(self.data, torch.Tensor):
            self.data = self.data.to(device, dtype, non_blocking)
        return self

    def to_csr(self) -> torch.Tensor:
        r"""
        Convert List to CSR Tensor.

        Notes:
        ------
        Each row in self.data should be the col indices !
        """
        crow_indices = np.cumsum([0] + list(map(len, self.data)), dtype=np.int64)
        col_indices = reduce(
            lambda x, y: x + y, self.data
        )
        values = np.ones_like(col_indices, dtype=np.int64)
        return torch.sparse_csr_tensor(
            crow_indices=crow_indices,
            col_indices=col_indices,
            values=values,
            size=(len(self.data), self.count) # B x Num of Items
        )


class FieldModule(Field, torch.nn.Module):
    r"""
    A module that represents a field of data.

    Attributes:
    -----------
    name: str 
        Name of this field.
    tags: set 
        A set of FieldTags.
    dtype: torch.dtype 
        Data type of the field.
    caster: Callable 
        Function to convert elements to specified dtype (int, float or str) with na_value.
    na_value: Union[str, int, float] 
        The fill value for null/missing values.
    dimension: int 
        The dimension of embeddings
    embeddings: torch.nn.Module
        Embedding module.

    Examples:
    ---------
    >>> from freerec.data.tags import USER, ITEM, ID, TARGET
    >>> User = SparseField('User', None, int, tags=(USER, ID))
    >>> Item = SparseField('Item', None, int, tags=(ITEM, ID))
    >>> Target = SparseField('Label', 0, int, transformer='none', tags=TARGET)
    """

    def __init__(
        self, name: str, na_value: Union[str, int, float], dtype: Callable,
        tags: Union[FieldTags, Iterable[FieldTags]] = tuple(),
        transformer: Union[str, Callable] = 'none'
    ) -> None:
        torch.nn.Module.__init__(self)
        Field.__init__(self, tags=tags)

        self.__name = name
        self.__na_value = na_value
        self.dtype = dtype

        self.transformer = TRANSFORMATIONS[transformer]() if isinstance(transformer, str) else transformer
        self.caster = partial(safe_cast, dest_type=dtype, default=na_value)

        self.dimension: int

    @property
    def name(self) -> str:
        return self.__name

    @property
    def na_value(self):
        return self.__na_value

    @property
    def dtype(self):
        return self.__dtype

    @dtype.setter
    def dtype(self, val):
        if val in (int, str):
            self.__dtype = torch.long
        elif val is float:
            self.__dtype = torch.float32
        else:
            self.__dtype = val

    def partial_fit(self, col) -> None:
        r"""
        Updates the transformer with a new batch of data.

        Parameters:
        -----------
        col: Iterable 
            The data to fit the transformer with.

        Returns:
        --------
        None
        """
        return self.transformer.partial_fit(col)

    def transform(self, col):
        r"""
        Applies the transformer to a new batch of data.

        Parameters:
        -----------
        col: Iterable
            The data to transform.

        Returns:
        --------
        The processed data.
        """
        return self.transformer.transform(col)

    def embed(self, dim: int, **kwargs):
        r"""
        Embed the field values into a lower dimensional space.

        Parameters:
        -----------
        dim: int 
            The dimension of the lower dimensional space.
        **kwargs: 
            Other arguments for embedding.

        Raises:
        -------
        NotImplementedError: if the method is not implemented.
        """
        raise NotImplementedError(f"{self.__class__.__name__}.embed() method should be implemented ...")

    def look_up(self, x: torch.Tensor) -> torch.Tensor:
        r"""
        Look up embeddings.

        Args:
        Parameters:
        -----------
        x: torch.Tensor 
            Indices for looking up.

        Raises:
        -------
        NotImplementedError: if the method is not implemented.
        """
        raise NotImplementedError(f"{self.__class__.__name__}.look_up() method should be implemented...")

    def __str__(self) -> str:
        tags = ','.join(map(str, self.tags))
        return f"{self.name}: [dtype: {self.dtype}, na_value: {self.na_value}, tags: {tags}]"


class SparseField(FieldModule):
    r""" 
    SparseField inherits from FieldModule and represents sparse features. 
    It is used to look up embeddings for categorical features.

    Attributes:
    -----------
    name: str 
        Name of this field.
    tags: set 
        A set of FieldTags.
    dtype: torch.dtype 
        Data type of the field.
    caster: Callable 
        Function to convert elements to specified dtype (int, float or str) with na_value.
    na_value: Union[str, int, float] 
        The fill value for null/missing values.
    dimension: int 
        The dimension of embeddings
    embeddings: torch.nn.Module
        Embedding module.

    Tags:
    -----
    SPARSE
    """

    def __init__(
        self, name: str, na_value: Union[str, int, float], dtype: Callable, 
        tags: Union[FieldTags, Iterable[FieldTags]] = tuple(),
        transformer: Union[str, Callable] = 'label2index'
    ):
        assert transformer == 'label2index', "SparseField supports 'label2index' only !"
        super().__init__(name, na_value, dtype, tags, transformer)
        self.add_tag(SPARSE)

    @property
    def count(self) -> Optional[int]:
        """Return the number of classes."""
        return self.transformer.count

    @property
    def enums(self) -> Optional[Tuple[int]]:
        """Return the tuple of IDs|..."""
        return self.transformer.enums

    def embed(self, dim: int, padding_idx: Optional[int] = None, **kwargs) -> None:
        r"""
        Create an nn.Embedding layer for the sparse field.

        Args:
        Parameters:
        -----------
        dim: int 
            The embedding dimension.
        **kwargs: 
            Other keyword arguments to be passed to nn.Embedding.

        Returns:
        --------
        None
        """
        self.dimension = dim
        nums = self.count if padding_idx is None else self.count + 1
        self.embeddings = torch.nn.Embedding(nums, dim, padding_idx=padding_idx, **kwargs)

    def look_up(self, x: torch.Tensor) -> torch.Tensor:
        r"""
        Look up embeddings for categorical features.

        Parameters:
        -----------
        x: torch.Tensor 
            The input tensor of shape (B, *), where B is the batch size
            and * can be any number of additional dimensions.

        Returns:
        --------
        embeddings: torch.Tensor 
            The output tensor of shape (B, *, d), where d is the embedding dimension.

        Examples:
        ---------
        >>> User: SparseField
        >>> ids = torch.arange(3).view(-1, 1)
        >>> User.look_up(ids).ndim
        3
        """
        return self.embeddings(x)


class DenseField(FieldModule):
    r"""
    DenseField is used for numerical features.

    Attributes:
    -----------
    name: str 
        Name of this field.
    tags: set 
        A set of FieldTags.
    dtype: torch.dtype 
        Data type of the field.
    caster: Callable 
        Function to convert elements to specified dtype (int, float or str) with na_value.
    na_value: Union[str, int, float] 
        The fill value for null/missing values.
    dimension: int 
        The dimension of embeddings
    embeddings: torch.nn.Module
        Embedding module.

    Tags:
    -----
    DENSE
    """

    def __init__(
        self, name: str, na_value: Union[str, int, float], dtype: Callable, 
        tags: Union[FieldTags, Iterable[FieldTags]] = tuple(),
        transformer: Union[str, Callable] = 'none'
    ):
        super().__init__(name, na_value, dtype, tags, transformer)
        self.add_tag(DENSE)

    def embed(self, dim: int = 1, bias: bool = False, linear: bool = False) -> None:
        r"""
        Create Embedding in nn.Linear manner.

        Parameters:
        -----------
        dim: int 
            Dimension.
        bias: bool 
            Add bias or not.
        linear: bool: 
            - `True`: nn.Linear.
            - `False`: using nn.Identity instead.
        **kwargs: other kwargs for nn.Linear.

        Returns:
        --------
        None.
        """
        if linear:
            self.dimension = dim
            self.embeddings = torch.nn.Linear(1, dim, bias=bias)
        else:
            self.dimension = 1
            self.embeddings = torch.nn.Identity()

    def look_up(self, x: torch.Tensor) -> torch.Tensor:
        r"""
        Look up embeddings for numeric features.
        Note that the return embeddings' shape is in accordance with
        that of categorical features.

        Parameters:
        -----------
            x: torch.Tensor in the shape of (B, *)

        Returns:
        --------
        embeddings: (B, *, 1) or (B, *, d)
            - If `linear` is True, it returns (B, *, d).
            - If `linear` is False, it returns (B, *, 1).

        Examples:
        ---------
        >>> Field: DenseField
        >>> vals = torch.rand(3, 1)
        >>> Field.look_up(vals).ndim
        3
        """
        return self.embeddings(x.unsqueeze(-1))
   

class FieldTuple(tuple):
    """A tuple of fields, which support attribute access and filtering by tags."""

    @lru_cache(maxsize=4)
    def groupby(self, *tags: FieldTags) -> 'FieldTuple':
        r"""
        Return those fields matching given tags.

        Parameters:
        -----------
        *tags: FieldTags 
            Variable length argument list of FieldTags to filter the fields by.

        Returns:
        --------
        A new FieldTuple that contains only the fields whose tags match the given tags.
        """
        return FieldTuple(field for field in self if field.match(*tags))

    @lru_cache(maxsize=4)
    def groupbynot(self, *tags: FieldTags) -> 'FieldTuple':
        r"""
        Return those fields not matching given tags.

        Parameters:
        -----------
        *tags: FieldTags 
            Variable length argument list of FieldTags to filter the fields by.

        Returns:
        --------
            A new FieldTuple that contains only the fields whose tags do not match the given tags.
        """
        return FieldTuple(field for field in self if not field.match(*tags))

    def state_dict(self) -> Dict:
        r"""
        Return state dict of fields.

        Returns:
        --------
            A dictionary containing the name and transformer of each field.
        """
        return {field.name: field.transformer for field in self}

    def load_state_dict(self, state_dict: Dict, strict: bool = False):
        r"""
        Load state dict of fields.

        Parameters:
        -----------
        state_dict: Dict 
            A dictionary containing the state of the fields.
        strict: bool 
            Whether to strictly enforce that the keys in the state dict match the names of the fields.
        """
        for field in self:
            field.transformer = state_dict.get(field.name, field.transformer)

    def copy(self) -> 'FieldTuple':
        r"""
        Return a copy of the FieldTuple.

        Returns:
        --------
        A new FieldTuple with the same fields as this one.
        """
        return FieldTuple(self)
    
    def index(self, *tags) -> int:
        r"""
        Get index by tags.

        Parameters:
        -----------
        *tags: FieldTags

        Returns:
        --------
        Index: int 
            Index of the field accordance with the given tags.
        
        Examples:
        ---------
        >>> from freerec.data.tags import USER, ITEM, ID
        >>> User = SparseField('User', None, int, tags=(USER, ID))
        >>> Item = SparseField('Item', None, int, tags=(ITEM, ID))
        >>> fields = FieldTuple([User, Item])
        >>> fields.index(USER, ID)
        0
        >>> fields.index(ITEM, ID)
        1
        """
        return super().index(self[tags])

    def __getitem__(self, index: Union[int, slice, FieldTags, Iterable[FieldTags]]) -> Union[Field, 'FieldTuple', None]:
        r"""
        Get fields by index.

        Parameters:
        -----------
        index: Union[int, slice, FieldTags, Iterable[FieldTags]]
            - int: Return the field at position `int`.
            - slice: Return the fields at positions of `slice`.
            - FieldTags: Return the fields matching `FieldTags`.
            - Iterable[FieldTags]: Return the fields matching `Iterable[FieldTags]`.

        Returns:
        --------
        Fields: Union[Field, FieldTuple, None]
            - Field: If only one field matches given tags.
            - None: If none of fields matches given tags.
            - FieldTuple: If more than one field match given tags.

        Examples:
        ---------
        >>> from freerec.data.tags import USER, ITEM, ID
        >>> User = SparseField('User', None, int, tags=(USER, ID))
        >>> Item = SparseField('Item', None, int, tags=(ITEM, ID))
        >>> fields = FieldTuple([User, Item])
        >>> fields[USER, ID] is User
        True
        >>> fields[0] is User
        True
        >>> fields[0:1] is User
        True
        >>> fields[Item, ID] is Item
        True
        >>> fields [1] is Item
        True
        >>> fields[1:] is Item
        True
        >>> len(fields[ID])
        2
        >>> len(fields[:])
        2
        >>> isinstance(fields[ID], FieldList)
        True
        """
        if isinstance(index, int):
            return super().__getitem__(index)
        elif isinstance(index, slice):
            fields = FieldTuple(
                super().__getitem__(index)
            )
        elif isinstance(index, FieldTags):
            fields =  self.groupby(index)
        else:
            fields = self.groupby(*index)
        if len(fields) == 1:
            return fields[0]
        elif len(fields) == 0:
            return None # for a safety purpose
        else:
            return fields


class FieldList(list):
    """A list of fields, which support attribute access and filtering by tags."""

    def groupby(self, *tags: FieldTags) -> 'FieldList':
        r"""
        Return those fields matching given tags.

        Parameters:
        -----------
        *tags: FieldTags 
            Variable length argument list of FieldTags to filter the fields by.

        Returns:
        --------
        A new FieldList that contains only the fields whose tags match the given tags.
        """
        return FieldList(field for field in self if field.match(*tags))

    def groupbynot(self, *tags: FieldTags) -> 'FieldList':
        r"""
        Return those fields not matching given tags.

        Parameters:
        -----------
        *tags: FieldTags 
            Variable length argument list of FieldTags to filter the fields by.

        Returns:
        --------
            A new FieldList that contains only the fields whose tags do not match the given tags.
        """
        return FieldList(field for field in self if not field.match(*tags))

    def copy(self) -> 'FieldList':
        r"""
        Return a copy of the FieldList.

        Returns:
        --------
        A new FieldList with the same fields as this one.
        """
        return FieldList(self)

    def index(self, *tags) -> int:
        r"""
        Get index by tags.

        Parameters:
        -----------
        *tags: FieldTags

        Returns:
        --------
        Index: int 
            Index of the field accordance with the given tags.
        
        Examples:
        ---------
        >>> from freerec.data.tags import USER, ITEM, ID
        >>> User = SparseField('User', None, int, tags=(USER, ID))
        >>> Item = SparseField('Item', None, int, tags=(ITEM, ID))
        >>> fields = FieldList([User, Item])
        >>> fields.index(USER, ID)
        0
        >>> fields.index(ITEM, ID)
        1
        """
        return super().index(self[tags])

    def __getitem__(self, index: Union[int, slice, FieldTags, Iterable[FieldTags]]) -> Union[Field, 'FieldList', None]:
        r"""
        Get fields by index.

        Parameters:
        -----------
        index: Union[int, FieldTags, Iterable[FieldTags]]
            - int: Return the field at position `int`.
            - slice: Return the fields at positions of `slice`.
            - FieldTags: Return the fields matching `FieldTags`.
            - Iterable[FieldTags]: Return the fields matching `Iterable[FieldTags]`.

        Returns:
        --------
        Fields: Union[Field, FieldList, None]
            - Field: If only one field matches given tags.
            - None: If none of fields matches given tags.
            - FieldList: If more than one field match given tags.

        Examples:
        ---------
        >>> from freerec.data.tags import USER, ITEM, ID
        >>> User = SparseField('User', None, int, tags=(USER, ID))
        >>> Item = SparseField('Item', None, int, tags=(ITEM, ID))
        >>> fields = FieldList([User, Item])
        >>> fields[USER, ID] is User
        True
        >>> fields[0] is User
        True
        >>> fields[0:1] is User
        True
        >>> fields[Item, ID] is Item
        True
        >>> fields [1] is Item
        True
        >>> fields[1:] is Item
        True
        >>> len(fields[ID])
        2
        >>> len(fields[:])
        2
        >>> isinstance(fields[ID], FieldList)
        True
        """

        if isinstance(index, int):
            return super().__getitem__(index)
        elif isinstance(index, slice):
            fields = FieldList(
                super().__getitem__(index)
            )
        elif isinstance(index, FieldTags):
            fields =  self.groupby(index)
        else:
            fields = self.groupby(*index)
        if len(fields) == 1:
            return fields[0]
        elif len(fields) == 0:
            return None # for a safety purpose
        else:
            return fields


class FieldModuleList(torch.nn.Module):
    r"""A collection of fields.
    
    Attributes:
        fields (nn.ModuleList): A list of fields.

    Examples:
        >>> from freerec.data.datasets import Gowalla_m1
        >>> basepipe = Gowalla_m1("../data")
        >>> fields = basepipe.fields
        >>> fields = FieldModule(fields)
    """

    def __init__(self, fields: Iterable[FieldModule]) -> None:
        super().__init__()

        self.fields = torch.nn.ModuleList(fields)

    @lru_cache(maxsize=4)
    def groupby(self, *tags: FieldTags) -> FieldList[FieldModule]:
        r"""
        Return those fields matching given tags.

        Parameters:
        -----------
        *tags: FieldTags 
            Variable length argument list of FieldTags to filter the fields by.

        Returns:
        --------
        A new FieldList that contains only the fields whose tags match the given tags.

        Examples:
        ---------
        >>> from freerec.data.tags import USER, ID
        >>> User = fields.groupby(USER, ID)
        >>> isinstance(User, List)
        True
        >>> Item = fields.groupby(ITEM, ID)[0]
        >>> Item.match(ITEM)
        True
        >>> Item.match(ID)
        True
        >>> Item.match(User)
        False
        """
        return FieldList(field for field in self.fields if field.match(*tags))

    @lru_cache(maxsize=4)
    def groupbynot(self, *tags: FieldTags) -> 'FieldList':
        r"""
        Return those fields not matching given tags.

        Parameters:
        -----------
        *tags: FieldTags 
            Variable length argument list of FieldTags to filter the fields by.

        Returns:
        --------
            A new FieldList that contains only the fields whose tags do not match the given tags.
        """
        return FieldList(field for field in self.fields if not field.match(*tags))

    def embed(self, dim: int, *tags: FieldTags, **kwargs):
        r"""
        Create embeddings.

        Parameters:
        -----------
        dim: int
            Dimension.
        *tags: FieldTags
        **kwargs: kwargs for nn.Embeddings or nn.Linear

        Examples:
        ---------
        >>> from freerec.data.tags import DENSE, SPARSE
        >>> fields.embed(8, SPARSE) # nn.Embedding(count, 8)
        >>> fields.embed(8, DENSE, bias=False) # nn.Linear(1, 8, bias=False)
        """
        for field in self.groupby(*tags):
            field.embed(dim, **kwargs)

    def calculate_dimension(self, *tags: FieldTags):
        """Return the total dimension of fields matching tiven tags."""
        return sum(field.dimension for field in self.groupby(*tags))

    def __len__(self):
        return len(self.fields)

    def __getitem__(self, index: Union[int, FieldTags, Iterable[FieldTags]]) -> Union[FieldModule, 'FieldList', None]:
        r"""
        Get fields by index.

        Parameters:
        -----------
        index: Union[int, FieldTags, Iterable[FieldTags]]
            - int: Return the field at position `int`.
            - slice: Return the fields at positions of `slice`.
            - FieldTags: Return the fields matching `FieldTags`.
            - Iterable[FieldTags]: Return the fields matching `Iterable[FieldTags]`.

        Returns:
        --------
        Fields: Union[FieldModule, FieldList, None]
            - Field: If only one field matches given tags.
            - None: If none of fields matches given tags.
            - FieldList: If more than one field match given tags.

        Examples:
        ---------
        >>> from freerec.data.tags import USER, ITEM, ID
        >>> User = SparseField('User', None, int, tags=(USER, ID))
        >>> Item = SparseField('Item', None, int, tags=(ITEM, ID))
        >>> fields = FieldModuleList([User, Item])
        >>> fields[USER, ID] is User
        True
        >>> fields[0] is User
        True
        >>> fields[0:1] is User
        True
        >>> fields[Item, ID] is Item
        True
        >>> fields [1] is Item
        True
        >>> fields[1:] is Item
        True
        >>> len(fields[ID])
        2
        >>> len(fields[:])
        2
        >>> isinstance(fields[ID], FieldList)
        True
        """

        if isinstance(index, int):
            return super().__getitem__(index)
        elif isinstance(index, slice):
            fields = FieldList(
                super().__getitem__(index)
            )
        elif isinstance(index, FieldTags):
            fields =  self.groupby(index)
        else:
            fields = self.groupby(*index)
        if len(fields) == 1:
            return fields[0]
        elif len(fields) == 0:
            return None # for a safety purpose
        else:
            return fields




from ..base import RecDataSet


__all__ = []


#TODO:===============================Session-based Recommendation===============================
class SessionBasedRecSet(RecDataSet):
    DATATYPE =  "Session"




from enum import Enum


class FieldTags(Enum):
    Sparse = 'Sparse'
    Dense = 'Dense'
    User = 'User'
    Item = 'Item'
    Timestamp = 'Timestamp'
    ID = 'ID'
    Feature = 'Feature'
    Target = 'Target'
    Positive = 'Positive'
    Negative = 'Negative'
    Seen = 'Seen'
    Unseen = 'Unseen'


SPARSE = FieldTags('Sparse')
DENSE = FieldTags('Dense')
USER = FieldTags('User')
ITEM = FieldTags('Item')
TIMESTAMP = FieldTags('Timestamp')
ID = FieldTags('ID')
FEATURE = FieldTags('Feature')
TARGET = FieldTags('Target')
POSITIVE = FieldTags('Positive')
NEGATIVE = FieldTags('Negative')
SEEN = FieldTags('Seen')
UNSEEN = FieldTags('Unseen')




import torchdata.datapipes as dp

from ..base import RecDataSet
from ...tags import TIMESTAMP, USER, ITEM, ID
from ...fields import Field, SparseField, DenseField
from ....utils import infoLogger
from ....dict2obj import Config


__all__ = ['SequentialRecSet']


class SequentialRecSet(RecDataSet):
    DATATYPE =  "Sequential"

    def check(self):
        assert isinstance(self.fields[TIMESTAMP], Field), "SequentialRecSet must have `TIMESTAMP' field."

    # def raw2pickle(self):
    #     """Convert raw data into (ordered) pickle format."""
    #     infoLogger(f"[{self.__class__.__name__}] >>> Convert raw data ({self.mode}) to chunks in pickle format")
    #     userIndex = self.fields.index(USER, ID)
    #     timeIndex = self.fields.index(TIMESTAMP)
    #     data = sorted(self.raw2data(), key=lambda row: (row[userIndex], row[timeIndex]))
    #     datapipe = dp.iter.IterableWrapper(data)

    #     count = 0
    #     for chunk in datapipe.batch(batch_size=self.DEFAULT_CHUNK_SIZE).column_():
    #         for j, field in enumerate(self.fields):
    #             chunk[j] = field.transform(chunk[j])
    #         self.write_pickle(chunk, count)
    #         count += 1
    #     infoLogger(f"[{self.__class__.__name__}] >>> {count} chunks done")


class UserItemTimeTriplet(SequentialRecSet):

    VALID_IS_TEST = False

    _cfg = Config(
        fields = [
            (SparseField, Config(name='UserID', na_value=None, dtype=int, tags=[USER, ID])),
            (SparseField, Config(name='ItemID', na_value=None, dtype=int, tags=[ITEM, ID])),
            (DenseField, Config(name='Timestamp', na_value=None, dtype=float, tags=[TIMESTAMP], transformer='none'))
        ]
    )

    open_kw = Config(mode='rt', delimiter='\t', skip_lines=1)

    def file_filter(self, filename: str):
        return self.mode in filename

    def raw2data(self) -> dp.iter.IterableWrapper:
        datapipe = dp.iter.FileLister(self.path)
        datapipe = datapipe.filter(filter_fn=self.file_filter)
        datapipe = datapipe.open_files(mode=self.open_kw.mode)
        datapipe = datapipe.parse_csv(delimiter=self.open_kw.delimiter, skip_lines=self.open_kw.skip_lines)
        datapipe = datapipe.map(self.row_processer)
        return datapipe

    def summary(self):
        super().summary()
        from prettytable import PrettyTable
        User, Item = self.fields[USER, ID], self.fields[ITEM, ID]

        table = PrettyTable(['#User', '#Item', '#Interactions', '#Train', '#Valid', '#Test', 'Density'])
        table.add_row([
            User.count, Item.count, self.trainsize + self.validsize + self.testsize,
            self.trainsize, self.validsize, self.testsize,
            (self.trainsize + self.validsize + self.testsize) / (User.count * Item.count)
        ])

        infoLogger(table)




import torchdata.datapipes as dp

from ..base import RecDataSet
from ...tags import USER, ITEM, ID
from ...fields import SparseField
from ....utils import infoLogger
from ....dict2obj import Config


__all__ = ['GeneralRecSet', 'UserItemPair']


class GeneralRecSet(RecDataSet):
    DATATYPE =  "General"


class UserItemPair(GeneralRecSet):
    r"""
    Implicit feedback data containing (User, Item) pairs.
    """

    VALID_IS_TEST = False

    _cfg = Config(
        fields = [
            (SparseField, Config(name='UserID', na_value=None, dtype=int, tags=[USER, ID])),
            (SparseField, Config(name='ItemID', na_value=None, dtype=int, tags=[ITEM, ID]))
        ]
    )

    open_kw = Config(mode='rt', delimiter='\t', skip_lines=1)

    def file_filter(self, filename: str):
        return self.mode in filename

    def raw2data(self) -> dp.iter.IterableWrapper:
        datapipe = dp.iter.FileLister(self.path)
        datapipe = datapipe.filter(filter_fn=self.file_filter)
        datapipe = datapipe.open_files(mode=self.open_kw.mode)
        datapipe = datapipe.parse_csv(delimiter=self.open_kw.delimiter, skip_lines=self.open_kw.skip_lines)
        datapipe = datapipe.map(self.row_processer)
        return datapipe

    def summary(self):
        super().summary()
        from prettytable import PrettyTable
        User, Item = self.fields[USER, ID], self.fields[ITEM, ID]

        table = PrettyTable(['#User', '#Item', '#Interactions', '#Train', '#Valid', '#Test', 'Density'])
        table.add_row([
            User.count, Item.count, self.trainsize + self.validsize + self.testsize,
            self.trainsize, self.validsize, self.testsize,
            (self.trainsize + self.validsize + self.testsize) / (User.count * Item.count)
        ])

        infoLogger(table)


class _Row2Pairer(dp.iter.IterDataPipe):

    def __init__(self, datapipe: dp.iter.IterDataPipe) -> None:
        super().__init__()
        self.source = datapipe

    def __iter__(self):
        for row in self.source:
            user = row[0]
            for item in row[1:]:
                if item:
                    yield user, item


class BARSUserItemPair(GeneralRecSet):
    r"""
    Implicit feedback data containing (User, Item) pairs.
    The data should be collected in the order of users; that is,
    each row represents a user's interacted items.
    """

    # Same validset and testset are the same now !
    VALID_IS_TEST = True

    _cfg = Config(
        fields = [
            (SparseField, Config(name='UserID', na_value=None, dtype=int, tags=[USER, ID])),
            (SparseField, Config(name='ItemID', na_value=None, dtype=int, tags=[ITEM, ID]))
        ]
    )

    open_kw = Config(mode='rt', delimiter=' ', skip_lines=0)

    def file_filter(self, filename: str):
        if self.mode == 'train':
            return 'train' in filename
        else:
            return 'test' in filename

    def raw2data(self) -> dp.iter.IterableWrapper:
        datapipe = dp.iter.FileLister(self.path)
        datapipe = datapipe.filter(filter_fn=self.file_filter)
        datapipe = datapipe.open_files(mode=self.open_kw.mode)
        datapipe = datapipe.parse_csv(delimiter=self.open_kw.delimiter, skip_lines=self.open_kw.skip_lines)
        datapipe = _Row2Pairer(datapipe)
        datapipe = datapipe.map(self.row_processer)
        return datapipe

    def summary(self):
        super().summary()
        from prettytable import PrettyTable
        User, Item = self.fields[USER, ID], self.fields[ITEM, ID]

        table = PrettyTable(['#User', '#Item', '#Interactions', '#Train', '#Test', 'Density'])
        table.add_row([
            User.count, Item.count, self.trainsize + self.testsize,
            self.trainsize, self.testsize,
            (self.trainsize + self.testsize) / (User.count * Item.count)
        ])

        infoLogger(table)




from typing import Iterator, List, Union, TypeVar, Any

import torch
import torchdata.datapipes as dp

from .base import Postprocessor
from ..fields import BufferField, FieldList, Field


__all__ = ["Columner", "ToTensor", "Fielder"]


T = TypeVar("T")


@dp.functional_datapipe("column_")
class Columner(Postprocessor):
    r"""
    A postprocessor that converts a batch of samples into columns.
    Columner takes a datapipe that yields a batch of samples, and converts them into columns. 
    This can be useful for the following transformation (by column).

    Parameters:
    -----------
    source_dp: dp.IterDataPipe 
        A datapipe that yields a batch samples.
    """

    def __iter__(self):
        for batch in self.source:
            yield list(zip(*batch))


@dp.functional_datapipe("tensor_")
class ToTensor(Postprocessor):
    r"""
    A datapipe that converts lists into torch Tensors.
    This class converts a List into a torch.Tensor.
    """

    def at_least_2d(self, vals: torch.Tensor):
        """Reshape tensor to 2D if needed."""
        return vals.unsqueeze(1) if vals.ndim == 1 else vals

    def to_tensor(self, col: List) -> Union[List, torch.Tensor]:
        """Convert the List to a torch.Tensor."""
        try:
            return self.at_least_2d(
                torch.tensor(col)
            )
        except ValueError: # avoid ragged List
            return col

    def __iter__(self) -> Iterator:
        for chunk in self.source:
            yield self.listmap(self.to_tensor, chunk)


@dp.functional_datapipe("field_")
class Fielder(Postprocessor):
    r"""
    Convert column data to field-style data.
    Then filtering by tags can be possible.
    """

    def __init__(
        self, source_dp: dp.iter.IterDataPipe, 
        *fields: Field
    ) -> None:
        super().__init__(source_dp, fields=fields)

    @staticmethod
    def _buffer(field: Field, col: Any):
        return field.buffer(col)

    def __iter__(self) -> Iterator[FieldList[BufferField]]:
        for cols in self.source:
            yield FieldList(map(
                self._buffer,
                self.fields,
                cols
            ))




from typing import Any, Callable, Iterable, List, Dict, Optional, Tuple, Union

import torch, abc, os, subprocess, shlex, time, sys
import pandas as pd
from torchdata.datapipes.iter import IterDataPipe
from torch.utils.tensorboard import SummaryWriter
from functools import partial
from itertools import product
from collections import defaultdict
from freeplot.utils import import_pickle, export_pickle

from .data.fields import FieldModule, FieldTuple
from .data.dataloader import DataLoader
from .models import RecSysArch
from .criterions import BaseCriterion
from .dict2obj import Config
from .utils import AverageMeter, Monitor, timemeter, infoLogger
from .metrics import *
from .parser import TIME


__all__ = ['ChiefCoach', 'Coach', 'Adapter']


DEFAULT_METRICS = {
    'LOSS': lambda x: x,
    #############
    'MSE': mean_squared_error,
    'MAE': mean_abs_error,
    'RMSE': root_mse,
    #############
    'PRECISION': precision,
    'RECALL': recall,
    'F1': f1_score,
    'AUC': auroc,
    'HITRATE': hit_rate,
    #############
    'NDCG': normalized_dcg,
    'MRR': mean_reciprocal_rank,
    'MAP': mean_average_precision
}

DEFAULT_FMTS = {
    'LOSS': ".5f",
    #############
    'MSE': ".4f",
    'MAE': ".4f",
    'RMSE': ".4f",
    #############
    'PRECISION': ".4f",
    'RECALL': ".4f",
    'F1': ".4f",
    'AUC': ".4f",
    'HITRATE': ".4f",
    #############
    'NDCG': ".4f",
    'MRR': ".4f",
    'MAP': ".4f",
}

DEFAULT_BEST_CASTER = {
    'LOSS': min,
    #############
    'MSE': min,
    'MAE': min,
    'RMSE': min,
    #############
    'PRECISION': max,
    'RECALL': max,
    'F1': max,
    'AUC': max,
    'HITRATE': max,
    #############
    'NDCG': max,
    'MRR': max,
    'MAP': max,
}


class _DummyModule(torch.nn.Module):
    """This is a dummy module that serves as a placeholder for a real model."""
    def forward(self, *args, **kwargs):
        """Dummy forward method that raises a `NotImplementedError`."""
        raise NotImplementedError("No model available for Coach ...")

    def step(self, *args, **kwargs):
        """Dummy step method that raises a `NotImplementedError`."""
        raise NotImplementedError("No optimizer or lr scheduler available for Coach ...")

    def backward(self, *args, **kwargs):
        """Dummy backward method that raises a `NotImplementedError`."""
        raise NotImplementedError("No optimizer available for Coach ...")


class ChiefCoach(metaclass=abc.ABCMeta):
    r""" 
    The `ChiefCoach` class is the top-level class for running the training and evaluation loops.

    Parameters:
    -----------
    trainpipe : IterDataPipe
        Iterable data pipeline for training data.
    validpipe : IterDataPipe, optional
        Iterable data pipeline for validation data.
        If `None`, use `trainpipe` instead.
    testpipe : IterDataPipe, optional
        Iterable data pipeline for testing data.
        If `None`, use `validpipe` instead.
    fields : Iterable[FieldModule]
        Tuple of `FieldModule`s for dataset fields.
    model : Union[RecSysArch, torch.nn.Module, None]
        Model for training and evaluating. 
        If `None`, use _DummyModule instead, which should not call `forward`.
    criterion : Union[BaseCriterion, Callable]
        Callable for computing the loss function.
    optimizer : torch.optim.Optimizer, optional
        Optimizer for updating model parameters. 
        If `None`, use _DummyModule instead, which should not call `step` and `backward`.
    lr_scheduler : torch.optim.lr_scheduler._LRScheduler, optional
        Learning rate scheduler. If `None`, use _DummyModule instead, 
        which should not call `step`.
    device : Union[torch.device, str, int]
        Device on which to run the computation. 
            - `torch.device`
            - `str`: Like `cpu`, `cuda:0`.
            - `int`: Using cuda:`int`.
    """


    def __init__(
        self, *,
        trainpipe: IterDataPipe, validpipe: Optional[IterDataPipe], testpipe: Optional[IterDataPipe], fields: Iterable[FieldModule],
        model: Union[RecSysArch, torch.nn.Module, None], criterion: Union[BaseCriterion, Callable], 
        optimizer: Optional[torch.optim.Optimizer], lr_scheduler: Optional[torch.optim.lr_scheduler._LRScheduler],
        device: Union[torch.device, str, int]
    ):

        self.fields: FieldTuple[FieldModule] = FieldTuple(fields)
        self.device = torch.device(device)

        self._set_datapipe(trainpipe, validpipe, testpipe)
        self._set_other(model, criterion, optimizer, lr_scheduler)

        self.__mode = 'train'

    def _set_datapipe(
        self,
        trainpipe,
        validpipe=None,
        testpipe=None,
    ):
        """Set the data pipe for training, validation and test."""
        self.trainpipe = trainpipe
        self.validpipe = self.trainpipe if validpipe is None else validpipe
        self.testpipe = self.validpipe if testpipe is None else testpipe

    def _set_other(
        self,
        model=None, criterion=None, optimizer=None, lr_scheduler=None,
    ):
        """Set the other necessary components."""
        self.criterion = criterion
        self.model = model.to(self.device) if model else _DummyModule()
        self.optimizer = optimizer if optimizer else _DummyModule()
        self.lr_scheduler = lr_scheduler if lr_scheduler else _DummyModule()

    @property
    def mode(self):
        """Get the current mode of the chief coach."""
        return self.__mode

    @timemeter("Coach/train")
    def train(self):
        """Start training and return the training loss."""
        self.__mode = 'train'
        self.model.train()
        return self.train_per_epoch()

    @timemeter("Coach/valid")
    @torch.no_grad()
    def valid(self):
        """Start validation and return the validation metrics."""
        self.__mode = 'valid'
        self.model.eval()
        return self.evaluate(prefix='valid')

    @timemeter("Coach/test")
    @torch.no_grad()
    def test(self):
        """Start testing and return the test metrics."""
        self.__mode = 'test'
        self.model.eval()
        return self.evaluate(prefix='test')

    @abc.abstractmethod
    def train_per_epoch(self):
        raise NotImplementedError(
            f"{self.__class__.__name__}.train_per_epoch() should be implemented ..."
        )

    @abc.abstractmethod
    def evaluate(self, prefix: str = 'valid'):
        raise NotImplementedError(
            f"{self.__class__.__name__}.evaluate() should be implemented ..."
        )

    def register_metric(
        self, name: str, func: Callable, 
        fmt: str = '.4f', best_caster: Callable = max
    ) -> None:
        r"""
        Register a metric.

        Parameters
        ----------
        name : str
            The complete name of the metric, such as `LOSS2`.
            The notation `@` should not be included, i.e., 'LOSS@2' is invalid.
        func : Callable
            The function to process the data for the metric.
        fmt : str, optional
            The format to use when printing the metric, defaults to `'.4f'`.
        best_caster : Callable, optional
            A function used to cast the best value of the metric, defaults to `max`.
            
        Returns
        -------
        None
        
        Raises
        ------
        AssertionError
            When `name` has already been registered or contains the notation `@`.
        """

        name = name.upper()
        assert DEFAULT_METRICS.get(name, None) is None, f"The metric {name} already exists ..."
        assert '@' not in name, f"The metric name has invalid notation of `@' ..."
        DEFAULT_METRICS[name] = func
        DEFAULT_FMTS[name] = fmt
        DEFAULT_BEST_CASTER[name] = best_caster


class Coach(ChiefCoach):
    """The framework for training."""

    def prepare_dataloader(self) -> None:
        """Prepare data loaders for training, validation, and testing data."""
        self.trainloader = DataLoader(
            datapipe=self.trainpipe, 
            num_workers=self.cfg.num_workers,
            pin_memory=self.cfg.pin_memory
        )
        self.validloader = DataLoader(
            datapipe=self.validpipe, 
            num_workers=self.cfg.num_workers,
            pin_memory=self.cfg.pin_memory
        )
        self.testloader = DataLoader(
            datapipe=self.testpipe, 
            num_workers=self.cfg.num_workers,
            pin_memory=self.cfg.pin_memory
        )
    
    @property
    def dataloader(self):
        """Return the corresponding data loader depending on the current mode."""
        if self.mode == 'train':
            return self.trainloader
        elif self.mode == 'valid':
            return self.validloader
        else:
            return self.testloader

    @property
    def monitors(self) -> Monitor:
        """Return the monitor dictionary for the different modes ('train', 'valid', 'test')."""
        return self.__monitors

    @property
    def meter4best(self):
        return self.__best_meter

    @meter4best.setter
    def meter4best(self, meter: AverageMeter):
        self.__best_meter = meter
        infoLogger(f"[Coach] >>> Set best meter: {meter.name} ")

    @timemeter("Coach/compile")
    def compile(
        self, cfg: Config, monitors: List[str], 
        which4best: str = 'LOSS'
    ):
        r"""
        Load the configuration and set up monitors for training.

        Parameters
        ----------
        cfg : Config
            A configuration object with the training details.
        monitors : List[str]
            A list of metric names to be monitored during training.
        which4best : str, defaults `LOSS'
            The metric used for selecting the best checkpoint.

        Examples
        --------
        >>> coach: Coach
        >>> coach.compile(cfg, monitors=['loss', 'recall@10', 'recall@20', 'ndcg@10', 'ndcg@20'])
        """
        self.cfg = cfg
        # meters for train|valid|test
        self.__monitors = Monitor()
        self.__monitors['train'] = defaultdict(list)
        self.__monitors['valid'] = defaultdict(list)
        self.__monitors['test'] = defaultdict(list)

        def set_monitor(
            name: str, lastname: str, prefix: str = 'train', **kwargs
        ):
            """Add a monitor for the specified metric."""
            try:
                meter = AverageMeter(
                        name=name,
                        metric=partial(DEFAULT_METRICS[lastname], **kwargs),
                        fmt=DEFAULT_FMTS[lastname],
                        best_caster=DEFAULT_BEST_CASTER[lastname]
                    )
                self.__monitors[prefix][lastname].append(meter)
            except KeyError:
                raise KeyError(
                    f"The metric of {lastname} is not included. "
                    f"You can register by calling `register_metric(...)' ..."
                )
            return meter

        # UPPER
        which4best = which4best.upper()
        monitors = ['LOSS'] + [name.upper() for name in monitors] + [which4best]
        monitors = sorted(set(monitors), key=monitors.index)

        for name in monitors:
            for prefix in ('train', 'valid', 'test'):
                if '@' in name:
                    lastname, K = name.split('@')
                    meter = set_monitor(
                        name=name,
                        lastname=lastname,
                        prefix=prefix,
                        k=int(K)
                    )
                else:
                    lastname = name
                    meter = set_monitor(
                        name=name,
                        lastname=lastname,
                        prefix=prefix
                    )
                if prefix == 'valid' and name == which4best:
                    self.meter4best = meter
                    self._best = -float('inf') if meter.caster is max else float('inf')
                    self._best_epoch = 0

        # Prepare data loaders
        self.prepare_dataloader()

    def save(self) -> None:
        """Save the model"""
        torch.save(self.model.state_dict(), os.path.join(self.cfg.LOG_PATH, self.cfg.SAVED_FILENAME))

    def load(self, path: str, filename: Optional[str] = None, **kwargs) -> None:
        filename = self.cfg.SAVED_FILENAME if filename is None else filename
        self.model.load_state_dict(torch.load(os.path.join(path, filename), **kwargs))

    def save_checkpoint(self, epoch: int) -> None:
        r"""
        Save current checkpoint at epoch.

        Parameters:
        -----------
            epoch :int Current epoch number.

        Returns:
        --------
            None
        """
        path = os.path.join(self.cfg.CHECKPOINT_PATH, self.cfg.CHECKPOINT_FILENAME)
        checkpoint = dict()
        checkpoint['epoch'] = epoch
        for module in self.cfg.CHECKPOINT_MODULES:
            checkpoint[module] = getattr(self, module).state_dict()
        checkpoint['monitors'] = self.monitors.state_dict()
        torch.save(checkpoint, path)

    def load_checkpoint(self) -> int:
        r"""
        Load last saved checkpoint.

        Returns:
        --------
        epoch: int 
            The epoch number loaded from the checkpoint.
        """
        path = os.path.join(self.cfg.CHECKPOINT_PATH, self.cfg.CHECKPOINT_FILENAME)
        checkpoint = torch.load(path)
        for module in self.cfg.CHECKPOINT_MODULES:
            getattr(self, module).load_state_dict(checkpoint[module])
        self.monitors.load_state_dict(checkpoint['monitors'])
        return checkpoint['epoch']

    def save_best(self) -> None:
        torch.save(self.model.state_dict(), os.path.join(self.cfg.LOG_PATH, self.cfg.BEST_FILENAME))

    def load_best(self) -> None:
        infoLogger(f"[Coach] >>> Load best model @Epoch {self._best_epoch:<4d} ")
        self.model.load_state_dict(torch.load(os.path.join(self.cfg.LOG_PATH, self.cfg.BEST_FILENAME)))

    def check_best(self, epoch: int) -> None:
        """Update best value."""
        if self.meter4best.active:
            best_ = self.meter4best.which_is_better(self._best)
            if best_ != self._best:
                self._best = best_
                self._best_epoch = epoch
                infoLogger(f"[Coach] >>> Better ***{self.meter4best.name}*** of ***{self._best:.4f}*** ")
                self.save_best()

    def eval_at_best(self):
        try:
            self.load_best()
            self.valid()
            self.test()
            self.step(self._best_epoch)
            self.load(self.cfg.LOG_PATH, self.cfg.SAVED_FILENAME)
        except FileNotFoundError:
            infoLogger(f"[Coach] >>> No best model was recorded. Skip it ...")

    def resume(self) -> int:
        r"""
        Resume training from the last checkpoint.

        Returns:
        --------
        start_epoch: int
            The epoch number to resume training from.
        """
        start_epoch: int = 0
        if self.cfg.resume:
            start_epoch = self.load_checkpoint()
            infoLogger(f"[Coach] >>> Load last checkpoint and train from epoch: {start_epoch}")
        return start_epoch

    @torch.no_grad()
    def monitor(
        self, *values,
        n: int = 1, mode: str = 'mean', 
        prefix: str = 'train', pool: Optional[Iterable] = None
    ):

        r"""
        Log data values to specific monitors.

        Parameters:
        -----------
        *values : data
            The data values to be logged.
        n : int
            The batch size in general.
        mode : str, optional
            The mode to compute the metric. Can be 'sum' or 'mean' (default).
        prefix : str, optional
            The prefix string indicating which mode the values belong to. Can be 'train', 'test' or 'valid'.
        pool : List[str], optional
            A list of metric names to log. If None, all metrics in the pool of `prefix` will be logged.
        """

        metrics: Dict[List] = self.monitors[prefix]
        pool = metrics if pool is None else pool
        for lastname in pool:
            for meter in metrics.get(lastname.upper(), []):
                meter(*values, n=n, mode=mode)

    def step(self, epoch: int):
        r"""
        Prints training status and evaluation results for each epoch, 
        and resets the corresponding `AverageMeter` instances.

        Parameters:
        -----------
        epoch : int
            The epoch number.

        Returns:
        --------
        None
        """
        metrics: Dict[str, List[AverageMeter]]
        for prefix, metrics in self.monitors.items():
            infos = [f"[Coach] >>> {prefix.upper():5} @Epoch: {epoch:<4d} >>> "]
            for meters in metrics.values():
                infos += [meter.step() for meter in meters if meter.active]
            infoLogger(' || '.join(infos))

    @timemeter("Coach/summary")
    def summary(self):
        r"""
        Summary the whole training process.

        Generate a summary of the entire training process, including the historical evaluation results, the best
        historical results, and the curves of historical results. The resulting summary is saved to a Markdown file named
        "Summary.md" in the `self.cfg.LOG_PATH` directory.

        Additionally, the best historical results are saved to a binary file named `self.cfg.MONITOR_BEST_FILENAME`.
        """
        s = "|  {prefix}  |   {name}   |   {val}   |   {epoch}   |   {img}   |\n"
        info = ""
        info += "|  Prefix  |   Metric   |   Best   |   @Epoch   |   Img   |\n"
        info += "| :-------: | :-------: | :-------: | :-------: | :-------: |\n"
        data = []
        best = defaultdict(dict)

        for prefix, metrics in self.monitors.items():
            metrics: defaultdict[str, List[AverageMeter]]
            freq = 1 if prefix == 'train' else self.cfg.EVAL_FREQ
            for lastname, meters in metrics.items():
                for meter in meters:
                    # Skip those meters never activated.
                    if len(meter.history) == 0:
                        continue
                    meter.plot(freq=freq)
                    imgname = meter.save(path=os.path.join(self.cfg.LOG_PATH, self.cfg.SUMMARY_DIR), prefix=prefix)
                    epoch, val = meter.argbest(freq)
                    info += s.format(
                        prefix=prefix, name=meter.name,
                        val=val, epoch=epoch, img=f"![]({imgname})"
                    )
                    data.append([prefix, meter.name, val, epoch])
                    if val != -1: # Only save available data.
                        best[prefix][meter.name] = val

        file_ = os.path.join(self.cfg.LOG_PATH, self.cfg.SUMMARY_DIR, self.cfg.SUMMARY_FILENAME)
        with open(file_, "w", encoding="utf8") as fh:
            fh.write(info)

        df = pd.DataFrame(data, columns=['Prefix', 'Metric', 'Best', '@Epoch'])
        infoLogger(str(df))
        infoLogger(f"[LoG_PaTH] >>> {self.cfg.LOG_PATH}")

        self.monitors.write(os.path.join(self.cfg.LOG_PATH, self.cfg.SUMMARY_DIR)) # tensorboard

        self.monitors.save(os.path.join(self.cfg.LOG_PATH, self.cfg.DATA_DIR), self.cfg.MONITOR_FILENAME)
        export_pickle(best, os.path.join(self.cfg.LOG_PATH, self.cfg.DATA_DIR, self.cfg.MONITOR_BEST_FILENAME))
    
    @timemeter("Coach/fit")
    def fit(self):
        start_epoch = self.resume()
        for epoch in range(start_epoch, self.cfg.epochs):
            if epoch % self.cfg.CHECKPOINT_FREQ == 0:
                self.save_checkpoint(epoch)
            if epoch % self.cfg.EVAL_FREQ == 0:
                if self.cfg.EVAL_VALID:
                    self.valid()
                if self.cfg.EVAL_TEST:
                    self.test()
            self.train()

            self.check_best(epoch)
            self.step(epoch)
        self.save()

        # last epoch
        self.valid()
        self.test()
        self.check_best(self.cfg.epochs)
        self.step(self.cfg.epochs)

        self.summary()

        self.eval_at_best()


class Adapter:
    r"""
    Params tuner.

    Flows:
    ------
    1. compile: configure the command, environments, and parameters for training.
    2. allocate devices for various parameters:
        - register the ID, log path, and device first
        - execute the command
        - collect information from the log path and output to TensorBoard
        - save the checkpoint
        - release the corresponding device

    Examples:
    ---------
    >>> cfg = {'command': 'python xxx.py', 'params': {'optimizer': ['sgd', 'adam']}}
    >>> tuner = Adapter()
    >>> tuner.compile(cfg)
    >>> tuner.fit()
    """

    def __init__(self) -> None:
        self.params = []
        self.values = []
        self.devices = []

    @property
    def COMMAND(self):
        return self.cfg.COMMAND

    def register(self, device: str) -> Tuple[str, str]:
        self.cfg.ENVS['id'] = time.strftime(TIME)
        self.cfg.ENVS['device'] = device
        command = self.COMMAND + self.get_option('id', self.cfg.ENVS.id)
        command += self.get_option('device', self.cfg.ENVS.device)
        return command, self.cfg.ENVS.id, self.cfg.LOG_PATH.format(**self.cfg.ENVS)

    @timemeter("Adapter/compile")
    def compile(self, cfg: Config) -> None:
        r"""
        Configure the command, environments, and parameters for training.

        Parameters:
        -----------
        cfg : Config
            An object that contains the command, environments, parameters, and defaults.

        Flows:
        ------
        1. Add environmental parameters to the basic `command`.
        2. Register all available devices.
        3. Convert all parameters from `cfg.PARAMS`.
        4. Convert all defaults from `cfg.DEFAULTS`.

        Returns:
        --------
        None
        """
        self.cfg = cfg
        piece = "\t{key}: {vals} \n"
        envs, params, defaults = "", "", ""
        for key, val in self.cfg.ENVS.items():
            if key == 'device':
                self.devices = val.split(',')
            else:
                self.cfg.COMMAND += self.get_option(key, val)
            envs += piece.format(key=key, vals=val)
        for key, vals in self.cfg.PARAMS.items():
            if isinstance(vals, (str, int, float)):
                vals = [vals]
            self.deploy_params(key, vals)
            params += piece.format(key=key, vals=vals)
        for key, val in self.cfg.DEFAULTS.items():
            self.cfg.DEFAULTS[key] = val
            defaults += piece.format(key=key, vals=val)

        cfg_infos = f"command: {self.cfg.COMMAND} \nenvs: \n{envs}params: \n{params}defaults: \n{defaults}"
        infoLogger(f"\033[0;31;47m{cfg_infos}\033[0m")
        

    def deploy_params(self, key: str, vals: Iterable):
        self.params.append(key)
        self.values.append(vals)

    @staticmethod
    def get_option(key: str, val: Any):
        r"""
        Convert (key, val) to '--key=val'.

        Parameters:
        -----------
        key : str
            The key of the parameter.
        val : Any
            The value of the parameter.

        Notes:
        ------
        All '_' in `key` will be replaced by '-'.

        Returns:
        --------
        str
            The parameter with format '--key=val'.

        Examples:
        ---------
        >>> Adapter.get_option('lr', '1e-3')
        '--lr=1e-3'
        >>> Adapter.get_option('learning_rate', '1e-3')
        '--learning-rate=1e-3'
        """
        return f" --{key.replace('_', '-')}={val}"

    def load_best(self, logPath: str):
        """Load best.pickle from logPath of corresponding."""
        file_ = os.path.join(logPath, self.cfg.DATA_DIR, self.cfg.MONITOR_BEST_FILENAME)
        return import_pickle(file_)

    def write(self, id_: str, logPath: str, params: Dict):
        r"""
        Write experiment results to tensorboard.

        Parameters:
        -----------
        id_: str
            Experiment ID.
        logPath: str
            Path to the experiment logs.
        params: Dict
            Configuration parameters of the experiment.

        Flows:
        ------
        1. Load the best data from `logPath`.
        2. Write the best data to tensorboard with `params`.

        Notes:
        ------
        If you find `-1` appearing in the tensorboard,
        it could mean that the data is of `str` type,
        which will cause an error if it is sent to tensorboard directly!
        """
        try:
            data = self.load_best(logPath)
            path = os.path.join(self.cfg.CORE_LOG_PATH, id_)
            with SummaryWriter(log_dir=path) as writer:
                metrics = dict()
                for prefix, best in data.items():
                    for metric, val in best.items():
                        val = val if isinstance(val, (int, float)) else -1
                        metrics['/'.join([prefix, metric])] = val
                writer.add_hparams(
                    params, metrics,
                )
        except Exception:
            infoLogger(
                f"\033[0;31;47m[Adapter] >>> Unknown errors happen. This is mainly due to abnormal exits of child processes.\033[0m"
            )


    def each_grid(self):
        """Grid search for each kind of param."""
        for key, vals in zip(self.params, self.values):
            for val in vals:
                yield self.cfg.DEFAULTS | {key: val}

    def product_grid(self):
        """Grid search across all combination of params"""
        for vals in product(*self.values):
            yield self.cfg.DEFAULTS | {option:val for option, val in zip(self.params, vals)}

    def save_checkpoint(self, source: List) -> None:
        """Save the rest of params."""
        path = os.path.join(self.cfg.CORE_CHECKPOINT_PATH, self.cfg.CHECKPOINT_FILENAME)
        checkpoint = dict()
        checkpoint['source'] = source
        torch.save(checkpoint, path)

    def load_checkpoint(self) -> int:
        """Load the rest of params."""
        path = os.path.join(self.cfg.CORE_CHECKPOINT_PATH, self.cfg.CHECKPOINT_FILENAME)
        checkpoint = torch.load(path)
        return checkpoint['source']

    @timemeter("Coach/resume")
    def resume(self):
        """Resume from the recent checkpoint."""
        source = self.each_grid() if self.cfg.EXCLUSIVE else self.product_grid()
        source = list(source)[::-1]
        source = self.load_checkpoint() if self.cfg.resume else source
        infoLogger(f"[Coach] >>> Load the recent checkpoint ...")
        return source

    def run(self, command: str, params: Dict):
        """Start a new subprocess"""
        for option, val in params.items():
            command += self.get_option(option, val)
        infoLogger(f"\033[0;31;47m{command}\033[0m")
        return subprocess.Popen(shlex.split(command))

    def wait(self, tasks: Dict):
        """Wait util all processes terminate."""
        for process_, id_, logPath, params in tasks.values():
            process_.wait()
            self.write(id_, logPath, params)

    def poll(self, tasks: Dict):
        """Wait util any process terminates."""
        buffer_source = []
        time.sleep(1) # for unique id
        while len(self.devices) == 0:
            time.sleep(7)
            for device, (process_, id_, logPath, params) in tasks.items():
                if process_.poll() is not None:
                    self.write(id_, logPath, params)
                    self.devices.append(device)
                else:
                    buffer_source.append(params)
        self.save_checkpoint(self.source + buffer_source)

    @timemeter("Adapter/fit")
    def fit(self):
        """Grid search."""
        self.source = self.resume()
        tasks = dict()
        try:
            while self.source:
                self.poll(tasks)
                params = self.source.pop()
                device = self.devices.pop()
                command, id_, logPath = self.register(device)
                process_ = self.run(command, params)
                tasks[device] = (process_, id_, logPath, params)
        except Exception as e:
            print(e)
        finally:
            self.wait(tasks)
            sys.exit()




from typing import Callable, Optional, Dict, List, Union
import torch
import numpy as np
from torch.utils.tensorboard import SummaryWriter

import logging, time, random, os
from collections import defaultdict
from freeplot.base import FreePlot
from freeplot.utils import export_pickle

from .dict2obj import Config


LOGGER = Config(
    name='RecSys', filename='log.txt', level=logging.DEBUG,
    filelevel=logging.DEBUG, consolelevel=logging.INFO,
    formatter=Config(
        filehandler=logging.Formatter('%(asctime)s:\t%(message)s'),
        consolehandler=logging.Formatter('%(message)s')
    ),
    info=print,
    debug=print
)

COLOR = {
    'current': "{0}",
    'cpu': "{0}",
    0: "{0}",
    1: "\033[1;35m{0}\033[0m",
    2: "\033[1;34m{0}\033[0m",
    3: "\033[1;33m{0}\033[0m",
    4: "\033[1;38m{0}\033[0m",
    5: "\033[1;32m{0}\033[0m"
}


class AverageMeter:
    r"""
    Computes and stores the average and current value of a metric.

    Parameters:
    -----------
    name: str
        The name of the meter.
    metric: Callable
        Metric function.
    fmt: str, optional (default: '.5f')
        Output format.
    best_caster: Callable, optional (default: max)
        The best caster between `min` or `max` based on the metric.
    """

    def __init__(
        self, name: str, metric: Callable, 
        fmt: str = ".5f", best_caster: Callable = max
    ):
        assert isinstance(metric, Callable), f"metric should be Callable but {type(metric)} received ..."
        self.name = name
        self.fmt = fmt
        self.caster = best_caster
        self.reset()
        self.__history = []
        self.__metric = metric

    @property
    def history(self):
        """Get the historical results."""
        return self.__history

    @history.setter
    def history(self, val: List):
        """Set the historical results."""
        self.__history = val.copy()

    def reset(self) -> None:
        """Reset the meter values"""
        self.val = 0.
        self.avg = 0.
        self.sum = 0.
        self.count = 0
        self.active = False

    def update(self, val: float, n: int = 1, mode: str = "mean") -> None:
        r"""
        Updates the meter.

        Parameters:
        ----------
        val: float
            Value.
        n: int, optional (default: 1)
            Batch size.
        mode: str, optional (default: "mean")
            Mode: 'sum'|'mean'.
        """
        self.val = val
        self.count += n
        if mode == "mean":
            self.sum += val * n
        elif mode == "sum":
            self.sum += val
        else:
            raise ValueError(f"Receive mode {mode} but 'mean' or 'sum' expected ...")
        self.avg = self.sum / self.count

    def step(self) -> str:
        r"""
        Saves the average value and resets the meter.

        Returns:
        --------
        info: str
            Average value in a formatted string.
        """
        self.history.append(self.avg)
        info = str(self)
        self.reset()
        return info

    def check(self, *values):
        r"""
        Check the metric value.

        Parameters:
        -----------
        values: variable-length argument list
            Metric function arguments.

        Returns:
        --------
        The metric value.
        """
        val = self.__metric(*values)
        if isinstance(val, torch.Tensor):
            val = val.cpu().numpy()
        else:
            val = np.array(val)
        if np.isnan(val) or np.isinf(val):
            raise ValueError(
                f"The metric of {self.name} got an unexpected value: {val.item()}."
            )
        return val.item()

    def plot(self, freq: int = 1) -> None:
        r"""
        Plots the meter values.
        
        Parameters:
        -----------
        freq: int, optional (default: 1)
            Plot frequency.
        """
        self.fp = FreePlot(
            shape=(1, 1),
            titles=(self.name,),
            dpi=300, latex=False
        )
        timeline = np.arange(len(self.history)) * freq
        self.fp.lineplot(timeline, self.history, marker='')
        self.fp.set_title(y=.98)

    def save(self, path: str, prefix: str = '') -> None:
        r"""
        Save the curves as a PNG file.

        Parameters:
        -----------
        path : str
            The path to save the file to.
        prefix : str, optional
            The prefix to add to the filename.

        Returns:
        --------
        filename : str
            The filename of the saved file.
        """
        filename = f"{prefix}{self.name}.png"
        self.fp.savefig(os.path.join(path, filename))
        return filename

    def which_is_better(self, other: float) -> bool:
        return self.caster(self.avg, other)

    def argbest(self, freq: int = 1) -> float:
        r"""
        Return the index and value of the best result in history.

        Parameters:
        -----------
        freq : int, optional
            The evaluation frequency, defaults to 1.

        Returns:
        --------
        index : float
            The index of the best result.
            If no available data, returns -1.
        value : float
            The value of the best result.
            If no available data, returns -1.
        """
        if len(self.history) == 0:
            return -1, -1
        indices = np.argsort(self.history)
        if self.caster is min:
            return indices[0] * freq, self.history[indices[0]]
        elif self.caster is max:
            return indices[-1] * freq, self.history[indices[-1]]
        else:
            raise ValueError("caster should be `min' or `max' ...")

    def __str__(self):
        r"""
        Return the string representation of this object.

        Returns:
        --------
        str
            The string representation of this object.
        """
        fmtstr = "{name} Avg: {avg:{fmt}}"
        return fmtstr.format(**self.__dict__)

    def __call__(self, *values, n: int = 1, mode: str = "mean")  -> None:
        r"""
        Add a new data point to the history.

        Parameters:
        -----------
        values : list or tuple
            The value(s) to add to the history.
        n : int, optional
            The number of times to add each value, defaults to 1.
        mode : str, optional
            The mode of adding values: 'mean' or 'sum', defaults to 'mean'.
        """
        self.active = True
        self.update(
            val = self.check(*values),
            n = n,
            mode = mode
        )


class Monitor(Config):

    def state_dict(self) -> Dict:
        r"""
        Return the state dictionary of monitors.

        Returns:
        --------
        state_dict : Dict
            The state dictionary that contains the history of all monitors.
        """
        state_dict = defaultdict(dict)
        monitors: Dict[str, List[AverageMeter]]
        for prefix, monitors in self.items():
            for metric, meters in monitors.items():
                state_dict[prefix][metric] = dict()
                for meter in meters:
                    state_dict[prefix][metric][meter.name] = meter.history
        return state_dict

    def load_state_dict(self, state_dict: Dict, strict: bool = False):
        r"""
        Load the history of monitors from the given state dictionary.

        Parameters:
        -----------
        state_dict : Dict
            The state dictionary that contains the history of monitors.
        strict : bool, optional (default=False)
            Whether to strictly enforce that the keys in the state dictionary match the keys in the monitor.
        """
        monitors: Dict[str, List[AverageMeter]]
        for prefix, monitors in self.items():
            for metric, meters in monitors.items():
                for meter in meters:
                    meter.history = state_dict[prefix][metric].get(meter.name, meter.history)

    def save(self, path: str, filename: str = 'monitors.pickle'):
        r"""
        Save the current state of the monitors to disk.

        Parameters:
        -----------
        path : str
            The path to the directory to save the state.
        filename : str, optional (default='monitors.pickle')
            The name of the file to save the state.
        """
        file_ = os.path.join(path, filename)
        export_pickle(self.state_dict(), file_)

    def write(self, path: str):
        r"""
        Write the history of monitors to Tensorboard.

        Parameters:
        -----------
        path : str
            The path to the directory to write the history to.
        """
        with SummaryWriter(path) as writer:
            monitors: Dict[str, List[AverageMeter]]
            for prefix, monitors in self.items():
                for metric, meters in monitors.items():
                    for meter in meters:
                        for t, val in enumerate(meter.history):
                            writer.add_scalar(
                                '/'.join([prefix, meter.name]),
                                val,
                                t
                            )


def set_logger(
    path: str,
    log2file: bool = True, log2console: bool = True
) -> None:
    r"""
    Set up a logger instance.

    Parameters:
    -----------
    path : str
        The path of the log file.
    log2file : bool, optional
        Whether to log messages to a file. Default is True.
    log2console : bool, optional
        Whether to log messages to console. Default is True.

    Returns:
    --------
    logger : logging.Logger
        A configured logger instance.
    """
    logger = logging.getLogger(LOGGER.name)
    logger.setLevel(LOGGER.level)

    if log2file:
        handler = logging.FileHandler(
            os.path.join(path, LOGGER.filename), 
            encoding='utf-8'
        )
        handler.setLevel(LOGGER.filelevel)
        handler.setFormatter(LOGGER.formatter.filehandler)
        logger.addHandler(handler)
    if log2console:
        handler = logging.StreamHandler()
        handler.setLevel(LOGGER.consolelevel)
        handler.setFormatter(LOGGER.formatter.consolehandler)
        logger.addHandler(handler)
    logger.debug("========================================================================")
    logger.debug("========================================================================")
    logger.debug("========================================================================")
    LOGGER['info'] = logger.info
    LOGGER['debug'] = logger.debug
    return logger

def set_color(device: Union[int, str]):
    r"""
    Set a color for the output of the specified device.

    Parameters:
    -----------
    device : Union[int, str]
        The device identifier, which can be an integer or string.
    """
    try:
        COLOR['current'] = COLOR[device]
    except KeyError:
        pass

def infoLogger(words: str):
    r"""
    Log an info-level message.

    Parameters:
    -----------
    words : str
        The message to log.

    Returns:
    --------
    words : str
        The message that was logged.
    """
    words = COLOR['current'].format(words)
    LOGGER.info(words)
    return words

def debugLogger(words: str):
    r"""
    Log a debug-level message.

    Parameters:
    -----------
    words : str
        The message to log.

    Returns:
    --------
    words : str
        The message that was logged.
    """
    LOGGER.debug(words)
    return words

def warnLogger(warn: str):
    r"""
    Log a warning-level message.

    Parameters:
    -----------
    warn : str
        The warning message to log.

    Returns:
    --------
    words : str
        The warning message that was logged.
    """
    words = f"\033[1;31m[Warning] >>> {warn} \033[0m"
    LOGGER.info(words)
    return words

def timemeter(prefix=""):
    r"""
    A decorator to measure the running time of a function.

    Parameters:
    -----------
    prefix : str, optional
        A prefix to be displayed in the logging message, by default "".

    Returns:
    --------
    wrapper : function
        The decorated function.
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            start = time.time()
            results = func(*args, **kwargs)
            end = time.time()
            infoLogger(f"[Wall TIME] >>> {prefix} takes {end-start:.6f} seconds ...")
            return  results
        wrapper.__doc__ = func.__doc__
        wrapper.__name__ = func.__name__
        return wrapper
    return decorator

def mkdirs(*paths: str) -> None:
    r"""
    Create directories.

    Parameters:
    -----------
    *paths : str
        Paths of directories to create.
    """
    for path in paths:
        try:
            os.makedirs(path)
        except FileExistsError:
            pass

def activate_benchmark(benchmark: bool) -> None:
    r"""
    Activate or deactivate the cudnn benchmark mode.

    Parameters:
    -----------
    benchmark : bool
        Whether to activate the benchmark mode.
    """
    from torch.backends import cudnn
    if benchmark:
        infoLogger(f"[Benchmark] >>> cudnn.benchmark == True | cudnn.deterministic == False")
        cudnn.benchmark, cudnn.deterministic = True, False
    else:
        infoLogger(f"[Benchmark] >>> cudnn.benchmark == False | cudnn.deterministic == True")
        cudnn.benchmark, cudnn.deterministic = False, True

def set_seed(seed: int) -> int:
    r"""
    Set the seed for the random number generators.

    Parameters:
    -----------
    seed : int
        The seed to set. If seed is -1, a random seed between 0 and 2048 will be generated.

    Returns:
    --------
    seed : int
        The actual seed used.
    """
    if seed == -1:
        seed = random.randint(0, 2048)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    infoLogger(f"[Seed] >>> Set seed: {seed}")
    return seed







from typing import Callable, Iterable, List, Tuple, Optional

import numpy as np
import os
import pandas as pd
import torchdata.datapipes as dp
from math import ceil, floor
from collections import defaultdict
from itertools import chain

from ..utils import mkdirs



def reporter(*attributes: str):
    def decorator(func: Callable):
        def wrapper(self, *args, **kwargs) -> Iterable:
            infos = [f"[{attr}: {getattr(self, attr)}]" for attr in attributes]
            infos = "  ".join(infos)
            print(f">>> Do `{func.__name__}' under the settings: {infos}")
            return func(self, *args, **kwargs)
        wrapper.__name__ = func.__name__
        wrapper.__doc__ = func.__doc__
        return wrapper
    return decorator

class Inter2Txt:

    def __init__(
        self,
        root: str, filename: str, dataset: str,
        kcore4user: int = 1, kcore4item: int = 1, star4pos: int = 0,
        ratios: Iterable = (8, 1, 1), strict: bool = True
    ) -> None:

        self.root = root
        self.filename = filename
        self.dataset = dataset
        self.kcore4user = int(kcore4user)
        self.kcore4item = int(kcore4item)
        self.star4pos = int(star4pos)
        self.ratios = ratios
        self.strict = strict

    @staticmethod
    def listmap(func: Callable, *args):
        return list(map(
            func, *args
        ))

    @staticmethod
    def where_is_(xxx: str, names: Iterable[str]):
        for i, name in enumerate(names):
            if name.startswith(xxx):
                return i
        return None
    
    @reporter('root', 'filename')
    def load_data(self):
        datapipe = dp.iter.FileLister(os.path.join(self.root, self.filename))
        datapipe = datapipe.filter(filter_fn=lambda file_: file_.endswith('.inter'))
        datapipe = datapipe.open_files(mode='rt')
        names = next(iter(
            datapipe.parse_csv(delimiter='\t', skip_lines=0)
        ))
        userIdx = self.where_is_('user', names)
        itemIdx = self.where_is_('item', names) # TODO: steam: product_id ?
        starIdx = self.where_is_('rating', names)
        timestampIdx = self.where_is_('timestamp', names)
        print(f"The .inter file includes columns: {names}")
        print(f"User@{userIdx} Item@{itemIdx} Timestamp@{timestampIdx} Rating@{starIdx}")

        datapipe = datapipe.parse_csv(delimiter='\t', skip_lines=1)
        if starIdx is None: # (User, Item, Time)
            datapipe = datapipe.map(
                lambda row: (str(row[userIdx]), str(row[itemIdx]), float(row[timestampIdx]))
            )
            data = list(datapipe)
        else: # (User, Item, Time, Star)
            datapipe = datapipe.map(
                lambda row: (str(row[userIdx]), str(row[itemIdx]), float(row[timestampIdx]), float(row[starIdx]))
            )
            data = self.filter_by_star(list(datapipe))
        return data

    @reporter('star4pos')
    def filter_by_star(self, data: Iterable) -> List:
        filtered = []
        for row in data:
            if row[3] >= self.star4pos:
                filtered.append(row[:3])
        return filtered

    @reporter('kcore4user', 'kcore4item', 'strict')
    def filter_by_core(self, data: Iterable) -> List:
        datasize = 0

        while datasize != len(data):
            datasize = len(data)
            print(f"datasize: {datasize}")

            count_per_user = defaultdict(int)
            count_per_item = defaultdict(int)
            users, items = set(), set()

            for row in data:
                user, item = row[0], row[1]
                count_per_user[user] += 1
                count_per_item[item] += 1
                if count_per_user[user] >= self.kcore4user:
                    users.add(user)
                if count_per_item[item] >= self.kcore4item:
                    items.add(item)
            data = list(filter(
                lambda row: row[0] in users and row[1] in items,
                data
            ))

            if not self.strict:
                break
        
        return data

    @reporter()
    def sort_by_timestamp(self, data: List) -> List:
        return sorted(data, key=lambda row: (row[0], row[2]))

    @reporter()
    def UI2ID(self, data: List) -> List:
        users, items = list(zip(*data))[:2]
        users, items = set(users), set(items)
        userMap = dict(zip(users, range(len(users))))
        itemMap = dict(zip(items, range(len(items))))
        self.userCount = len(users)
        self.itemCount = len(items)

        data = self.listmap(
            lambda row: (userMap[row[0]], itemMap[row[1]], row[2]),
            data
        )

        return data

    @reporter()
    def user_item_only(self, data: List) -> List:
        return self.listmap(lambda row: (row[0], row[1]), data)

    @reporter()
    def user_item_time_only(self, data: List) -> List:
        return self.listmap(lambda row: (row[0], row[1], row[2]), data)

    @reporter()
    def group_by_user(self, data: Iterable) -> List:
        data_by_user = defaultdict(list)
        for row in data:
            data_by_user[row[0]].append(row) # (User, Item, TimeStamp)
        return data_by_user

    def summary(self):
        from prettytable import PrettyTable
        table = PrettyTable(['#User', '#Item', '#Interactions', '#Train', '#Valid', '#Test', 'Density'])
        table.add_row([
            self.userCount, self.itemCount, self.trainsize + self.validsize + self.testsize,
            self.trainsize, self.validsize, self.testsize,
            (self.trainsize + self.validsize + self.testsize) / (self.userCount * self.itemCount)
        ])
        print(table)


class GenInter2Txt(Inter2Txt):

    @reporter('ratios')
    def split_by_ratio(self, data_by_user: List) -> Tuple[List, List, List]:
        trainset = []
        validset = []
        testset = []
        markers = np.cumsum(self.ratios)
        for user in range(self.userCount):
            pairs = data_by_user[user]
            if len(pairs) == 0:
                continue
            l = max(floor(markers[0] * len(pairs) / markers[-1]), 1)
            r = floor(markers[1] * len(pairs) / markers[-1])
            trainset.append(pairs[:l])
            if l < r:
                validset.append(pairs[l:r])
            if r < len(pairs):
                testset.append(pairs[r:])
        trainset = list(chain(*trainset))
        validset = list(chain(*validset))
        testset = list(chain(*testset))

        self.trainsize = len(trainset)
        self.validsize = len(validset)
        self.testsize = len(testset) 
        return trainset, validset, testset

    @reporter('root')
    def save(self, trainset, validset, testset):
        code = f"{self.kcore4user}{self.kcore4item}{self.star4pos}{''.join(map(str, self.ratios))}"
        path = os.path.join(
            self.root, 'General', 
            '_'.join([self.dataset, code, 'Chron'])
        )
        mkdirs(path)

        df = pd.DataFrame(trainset, columns=['User', 'Item'])
        df.to_csv(os.path.join(path, 'train.txt'), sep='\t', index=False)

        df = pd.DataFrame(validset, columns=['User', 'Item'])
        df.to_csv(os.path.join(path, 'valid.txt'), sep='\t', index=False)

        df = pd.DataFrame(testset, columns=['User', 'Item'])
        df.to_csv(os.path.join(path, 'test.txt'), sep='\t', index=False)

    def compile(self):
        data = self.load_data()
        data = self.filter_by_core(data)
        data = self.UI2ID(data)
        data = self.sort_by_timestamp(data)
        data = self.user_item_only(data)
        data = self.group_by_user(data)
        self.save(*self.split_by_ratio(data))
        self.summary()
    

class SeqInter2Txt(Inter2Txt):

    @reporter()
    def split(self, data_by_user: List) -> Tuple[List, List, List]:
        trainset = []
        validset = []
        testset = []
        for user in range(self.userCount):
            pairs = data_by_user[user]
            if len(pairs) == 0:
                continue
            if len(pairs) <= 3:
                trainset.append(pairs)
            else:
                trainset.append(pairs[:-2])
                validset.append(pairs[-2:-1])
                testset.append(pairs[-1:])

        trainset = list(chain(*trainset))
        validset = list(chain(*validset))
        testset = list(chain(*testset))

        self.trainsize = len(trainset)
        self.validsize = len(validset)
        self.testsize = len(testset) 
        return trainset, validset, testset

    @reporter('root')
    def save(self, trainset, validset, testset):
        code = f"{self.kcore4user}{self.kcore4item}{self.star4pos}"
        path = os.path.join(
            self.root, 'Sequential', 
            '_'.join([self.dataset, code, 'Chron'])
        )
        mkdirs(path)

        df = pd.DataFrame(trainset, columns=['User', 'Item', 'Timestamp'])
        df.to_csv(os.path.join(path, 'train.txt'), sep='\t', index=False)

        df = pd.DataFrame(validset, columns=['User', 'Item', 'Timestamp'])
        df.to_csv(os.path.join(path, 'valid.txt'), sep='\t', index=False)

        df = pd.DataFrame(testset, columns=['User', 'Item', 'Timestamp'])
        df.to_csv(os.path.join(path, 'test.txt'), sep='\t', index=False)

    def compile(self):
        data = self.load_data()
        data = self.filter_by_core(data)
        data = self.UI2ID(data)
        data = self.sort_by_timestamp(data)
        data = self.user_item_time_only(data)
        data = self.group_by_user(data)
        self.save(*self.split(data))
        self.summary()




from typing import List, Tuple

import random
import torchdata.datapipes as dp

from .base import Postprocessor
from ..datasets.base import RecDataSet
from ..fields import SparseField
from ..tags import USER, ITEM, ID
from ...utils import timemeter


__all__ = [
    'GenTrainUniformSampler', 'GenValidYielder', 'GenTestYielder',
    'SeqTrainYielder', 'SeqValidYielder', 'SeqTestYielder', 
    'SeqTrainUniformSampler', 'SeqValidSampler', 'SeqTestSampler'
]


#===============================For General Recommendation===============================

@dp.functional_datapipe("gen_train_uniform_sampling_")
class GenTrainUniformSampler(Postprocessor):
    r"""
    A functional datapipe for uniformly sampling users and their negatives.

    Parameters:
    -----------
    source_dp: dp.iter.IterableWrapper 
        A datapipe that yields users.
    dataset: RecDataSet 
        The dataset object that contains field objects.
    num_negatives: int 
        The number of negative samples for each piece of data.  
    """

    def __init__(
        self, source_dp: dp.iter.IterableWrapper,
        dataset: RecDataSet,
        num_negatives: int = 1,
    ) -> None:
        super().__init__(source_dp)
        self.num_negatives = num_negatives
        self.User: SparseField = dataset.fields[USER, ID]
        self.Item: SparseField = dataset.fields[ITEM, ID]
        self.prepare(dataset)

    @timemeter("GenUniformSampler/prepare")
    def prepare(self, dataset: RecDataSet):
        r"""
        Prepare the data before sampling.

        Parameters:
        -----------
        dataset: RecDataSet 
            The dataset object that contains field objects.
        """
        self.posItems = [set() for _ in range(self.User.count)]
        self.negative_pool = self._sample_from_all(dataset.datasize * self.num_negatives)
        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.posItems[user].add(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        self.posItems = [tuple(items) for items in self.posItems]

    def _check(self, user) -> bool:
        return len(self.posItems[user]) > 0

    def _sample_from_all(self, pool_size: int = 51200):
        r"""
        Randomly sample items from all items.

        Parameters:
        -----------
        pool_size: int 
            The number of items to be sampled.

        Returns:
        --------
        Generator: A generator that yields sampled items.
        """
        allItems = self.Item.enums
        while 1:
            for item in random.choices(allItems, k=pool_size):
                yield item

    def _sample_from_pool(self, seen: Tuple):
        r"""
        Randomly sample a negative item from the pool of all items.

        Parameters:
        -----------
        seen: set 
            A set of seen items.

        Returns:
        --------
        negative: int 
            A negative item that has not been seen.
        """
        negative = next(self.negative_pool)
        while negative in seen:
            negative = next(self.negative_pool)
        return negative

    def _sample_pos(self, user: int) -> int:
        r"""
        Randomly sample a positive item for a user.

        Parameters:
        -----------
        user: int 
            A user index.

        Returns:
        --------
        positive: int 
            A positive item that the user has interacted with.
        """
        return random.choice(self.posItems[user])

    def _sample_neg(self, user: int) -> List[int]:
        r"""Randomly sample negative items for a user.

        Parameters:
        ----------
        user: int 
            A user index.

        Returns:
        --------
        negatives: List[int] 
            A list of negative items that the user has not interacted with.
        """
        seen = self.posItems[user]
        return self.listmap(self._sample_from_pool, [seen] * self.num_negatives)

    def __iter__(self):
        for user in self.source:
            if self._check(user):
                yield [user, self._sample_pos(user), self._sample_neg(user)]


@dp.functional_datapipe("gen_valid_yielding_")
class GenValidYielder(Postprocessor):
    r"""
    A datapipe that yields (user, unseen, seen) triplets.
    The ValidTripleter postprocessor takes a RecDataSet as input,
    and yields (user, unseen, seen) triplets. The unseen and seen sets contain the IDs of
    the items that the user has not seen and seen, respectively. Whether to use validation
    or test set as the source of unseen items depends on the value of `dataset.VALID_IS_TEST`.
    """

    def __init__(
        self, source_dp: dp.iter.IterDataPipe,
        dataset: RecDataSet
    ) -> None:
        r"""
        Initializes a new instance of the Tripleter postprocessor.
        
        Parameters:
        -----------
        source_dp: RecDatapipe or Postprocessor 
            A RecDatapipe or another Postprocessor that yields dicts of NumPy arrays.
        dataset: RecDataSet 
            The dataset that provides the data source.
        """
        super().__init__(source_dp)

        self.User: SparseField = dataset.fields[USER, ID]
        self.Item: SparseField = dataset.fields[ITEM, ID]

        self.prepare(dataset)

    @timemeter("GenValidYielder/prepare")
    def prepare(self, dataset: RecDataSet):
        r"""
        Prepares the dataset by building sets of seen items for each user.

        Parameters:
        -----------
        dataset: RecDataSet 
            The dataset that provides the data source.
        """
        self.seenItems = [set() for _ in range(self.User.count)]
        self.unseenItems = [set() for _ in range(self.User.count)]

        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.seenItems[user].add(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        for chunk in dataset.valid():
            self.listmap(
                lambda user, item: self.unseenItems[user].add(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )

        self.seenItems = [tuple(items) for items in self.seenItems]
        self.unseenItems = [tuple(items) for items in self.unseenItems]

    def _check(self, user) -> bool:
        return len(self.unseenItems[user]) > 0

    def __iter__(self):
        r"""
        Yields:
        -------
        user, unseen, seen: int, List[int], List[int]
            Triplets for each user in the data source.
        """
        for user in self.source:
            if self._check(user):
                yield [user, self.unseenItems[user], self.seenItems[user]]


@dp.functional_datapipe("gen_test_yielding_")
class GenTestYielder(GenValidYielder):
    r"""
    A datapipe that yields (user, unseen, seen) triplets from the test set.
    The TestTriplet postprocessor takes a RecDataSet as input,
    and yields (user, unseen, seen) triplets from the test set. The unseen and seen sets contain the IDs of
    the items that the user has not seen and seen, respectively.
    """

    @timemeter("GenTestYielder/prepare")
    def prepare(self, dataset: RecDataSet):
        r"""
        Prepares the dataset by building sets of seen items for each user.

        Parameters:
        -----------
        dataset: RecDataSet 
            The dataset that provides the data source.
        """
        self.seenItems = [set() for _ in range(self.User.count)]
        self.unseenItems = [set() for _ in range(self.User.count)]

        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.seenItems[user].add(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        for chunk in dataset.valid():
            self.listmap(
                lambda user, item: self.seenItems[user].add(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        for chunk in dataset.test():
            self.listmap(
                lambda user, item: self.unseenItems[user].add(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )

        self.seenItems = [tuple(items) for items in self.seenItems]
        self.unseenItems = [tuple(items) for items in self.unseenItems]


#===============================For Sequential Recommendation===============================


@dp.functional_datapipe("seq_train_yielding_")
class SeqTrainYielder(Postprocessor):
    r"""
    A functional datapipe for yielding (user, positives, targets).

    Parameters:
    -----------
    source_dp: dp.iter.IterableWrapper 
        A datapipe that yields users.
    dataset: RecDataSet 
        The dataset object that contains field objects.
    """

    def __init__(
        self, source_dp: dp.iter.IterableWrapper,
        dataset: RecDataSet,
    ) -> None:
        super().__init__(source_dp)
        self.User: SparseField = dataset.fields[USER, ID]
        self.Item: SparseField = dataset.fields[ITEM, ID]
        self.prepare(dataset)

    @timemeter("SeqTrainYielder/prepare")
    def prepare(self, dataset: RecDataSet):
        r"""
        Prepare the data before yielding.

        Parameters:
        -----------
        dataset: RecDataSet 
            The dataset object that contains field objects.
        """
        self.posItems = [[] for _ in range(self.User.count)]
        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        self.posItems = [tuple(items) for items in self.posItems]

    def _check(self, user) -> bool:
        return len(self.posItems[user]) > 1

    def __iter__(self):
        for user in self.source:
            if self._check(user):
                posItems = self.posItems[user]
                yield [user, posItems[:-1], posItems[1:]]


@dp.functional_datapipe("seq_valid_yielding_")
class SeqValidYielder(SeqTrainYielder):

    @timemeter("SeqValidYielder/prepare")
    def prepare(self, dataset: RecDataSet):
        r"""
        Prepare the data before yielding.

        Parameters:
        -----------
        dataset: RecDataSet 
            The dataset object that contains field objects.
        """
        self.posItems = [[] for _ in range(self.User.count)]
        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        for chunk in dataset.valid():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        self.posItems = [tuple(items) for items in self.posItems]


@dp.functional_datapipe("seq_test_yielding_")
class SeqTestYielder(SeqValidYielder):
    r"""
    A functional datapipe for yielding (user, positives, targets).

    Parameters:
    -----------
    source_dp: dp.iter.IterableWrapper 
        A datapipe that yields users.
    dataset: RecDataSet 
        The dataset object that contains field objects.
    """

    @timemeter("SeqTestYielder/prepare")
    def prepare(self, dataset: RecDataSet):
        r"""
        Prepare the data before yielding.

        Parameters:
        -----------
        dataset: RecDataSet 
            The dataset object that contains field objects.
        """
        self.posItems = [[] for _ in range(self.User.count)]
        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        for chunk in dataset.valid():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        for chunk in dataset.test():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        self.posItems = [tuple(items) for items in self.posItems]


@dp.functional_datapipe("seq_train_uniform_sampling_")
class SeqTrainUniformSampler(SeqTrainYielder):


    @timemeter("SeqTrainUniformSampler/prepare")
    def prepare(self, dataset: RecDataSet):
        r"""
        Prepare the data before sampling.

        Parameters:
        -----------
        dataset: RecDataSet
            The dataset object that contains field objects.
        """
        self.posItems = [[] for _ in range(self.User.count)]
        self.negative_pool = self._sample_from_all(dataset.datasize)

        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )

        self.posItems = [tuple(items) for items in self.posItems]


    def _sample_from_all(self, pool_size: int = 51200):
        r"""
        Randomly sample items from all items.

        Parameters:
        -----------
        pool_size: int 
            The number of items to be sampled.

        Returns:
        --------
        Generator: A generator that yields sampled items.
        """
        allItems = self.Item.enums
        while 1:
            for item in random.choices(allItems, k=pool_size):
                yield item

    def _sample_from_pool(self, seen: Tuple):
        r"""
        Randomly sample a negative item from the pool of all items.

        Parameters:
        -----------
        seen: set 
            A set of seen items.

        Returns:
        --------
        negative: int 
            A negative item that has not been seen.
        """
        negative = next(self.negative_pool)
        while negative in seen:
            negative = next(self.negative_pool)
        return negative

    def _sample_neg(self, user: int) -> List[int]:
        r"""Randomly sample negative items for a user.

        Parameters:
        ----------
        user: int 
            A user index.

        Returns:
        --------
        negatives: List[int] 
            A list of negative items that the user has not interacted with.
        """
        seen = self.posItems[user]
        return self.listmap(self._sample_from_pool, [seen] * (len(seen) - 1))

    def __iter__(self):
        for user in self.source:
            if self._check(user):
                posItems = self.posItems[user]
                yield [user, posItems[:-1], posItems[1:], self._sample_neg(user)]


@dp.functional_datapipe("seq_valid_sampling_")
class SeqValidSampler(SeqValidYielder):


    def _sample_negs(self, seen: List[int]):
        unseen = list(set(self.Item.enums) - set(seen))
        return tuple(random.choices(unseen, k=100))

    @timemeter("SeqValidSampler/prepare")
    def prepare(self, dataset: RecDataSet):
        r"""
        Prepare the data before sampling.

        Parameters:
        -----------
        dataset: RecDataSet
            The dataset object that contains field objects.
        """
        self.posItems = [[] for _ in range(self.User.count)]
        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )

        self.negItems = self.listmap(
            self._sample_negs, self.posItems
        )

        for chunk in dataset.valid():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        self.posItems = [tuple(items) for items in self.posItems]

    def __iter__(self):
        for user in self.source:
            if self._check(user):
                posItems = self.posItems[user]
                yield [user, posItems[:-1], posItems[-1:] + self.negItems[user]]


@dp.functional_datapipe("seq_test_sampling_")
class SeqTestSampler(SeqValidSampler):

    @timemeter("SeqTestSampler/prepare")
    def prepare(self, dataset: RecDataSet):
        r"""
        Prepare the data before sampling.

        Parameters:
        -----------
        dataset: RecDataSet
            The dataset object that contains field objects.
        """
        self.posItems = [[] for _ in range(self.User.count)]
        for chunk in dataset.train():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        for chunk in dataset.valid():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )

        self.negItems = self.listmap(
            self._sample_negs, self.posItems
        )

        for chunk in dataset.test():
            self.listmap(
                lambda user, item: self.posItems[user].append(item),
                chunk[USER, ID], chunk[ITEM, ID]
            )
        self.posItems = [tuple(items) for items in self.posItems]




import torch
# import torchdata.datapipes as dp
# from torchdata.dataloader2 import DataLoader2, MultiProcessingReadingService


def _collate_fn(batch):
    return batch[0]

class DataLoader(torch.utils.data.DataLoader):

    def __init__(self, datapipe, num_workers: int = 0, **kwargs) -> None:
        super().__init__(
            dataset=datapipe, num_workers=num_workers, 
            batch_size=1, collate_fn=_collate_fn,
            **kwargs
        )


# TODO: torchdata==0.4.1 has the following issue:
# Multiprocessing and batching / collation:
# https://github.com/pytorch/data/issues/530
# class DataLoader(DataLoader2):

#     def __init__(
#         self, datapipe: dp.iter.IterDataPipe, 
#         num_workers: int = 0, **kwargs
#     ) -> None:
#         super().__init__(
#             datapipe=datapipe,
#             reading_service=MultiProcessingReadingService(
#                 num_workers=num_workers, **kwargs
#             )
#         )




from .base import *
from .general.base import *
from .sequential.base import *
from .session.base import *
from .context.base import *
from .knowledge.base import *

from . import general, sequential, session, context, knowledge




from ..base import RecDataSet


__all__ = []


#TODO:===============================Knowledge-based Recommendation===============================
class KnowledgeBasedRecSet(RecDataSet):
    DATATYPE =  "Knowledge"




from typing import Optional, Union, overload, Tuple
import torch
import torch.nn as nn


__all__ = ['RecSysArch']


class RecSysArch(nn.Module):
    r"""
    A PyTorch Module for recommendation system architecture.
    This module contains methods for broadcasting tensors and initializing the module parameters.

    Parameters:
    -----------
    nn.Module : PyTorch Module
        The base class of this module.

    Methods:
    --------
    to(device, dtype, non_blocking)
        Moves and/or casts the parameters and buffers.

    initialize()
        Initializes the module parameters.

    broadcast(*tensors)
        Broadcasts the given tensors according to broadcasting semantics.

    """

    def to(
        self, device: Optional[Union[int, torch.device]] = None,
        dtype: Optional[Union[torch.dtype, str]] = None,
        non_blocking: bool = False
    ):
        r"""
        Moves and/or casts the parameters and buffers.

        Parameters:
        -----------
        device : Union[int, torch.device], optional
            The destination device of the parameters and buffers, by default None.
        dtype : Union[torch.dtype, str], optional
            The desired data type of the parameters and buffers, by default None.
        non_blocking : bool, optional
            Whether the copy should be asynchronous or not, by default False.

        Returns:
        --------
        nn.Module
            The module with parameters and buffers moved and/or cast.
        """
        if device:
            self.device = device
        return super().to(device, dtype, non_blocking)

    def initialize(self):
        """Initializes the module parameters."""
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.xavier_normal_(m.weight)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0.)
            elif isinstance(m, nn.Embedding):
                nn.init.normal_(m.weight, std=1e-4)
            elif isinstance(m, (nn.BatchNorm1d, nn.BatchNorm2d)):
                nn.init.constant_(m.weight, 1.)
                nn.init.constant_(m.bias, 0.)


    @staticmethod
    def broadcast(*tensors: torch.Tensor):
        r"""
        Broadcasts the given tensors according to Broadcasting semantics.
        See [here](https://pytorch.org/docs/stable/generated/torch.broadcast_tensors.html#torch.broadcast_tensors) for details.

        Parameters:
        -----------
        tensors : torch.Tensor
            The tensors to broadcast.

        Returns:
        --------
        Tuple[torch.Tensor, ...]
            The broadcasted tensors.

        Examples:
        ---------
        >>> users = torch.rand(4, 1, 4)
        >>> items = torch.rand(4, 2, 1)
        >>> users, items = RecSysArch.broadcast(users, items)
        >>> users.size()
        torch.Size(4, 2, 4)
        >>> items.size()
        torch.Size(4, 2, 4)
        """
        return torch.broadcast_tensors(*tensors)


    @overload
    def recommand(self) -> Tuple[torch.Tensor, torch.Tensor]: ...

    @overload
    def recommand(self, seqs: torch.Tensor, items: torch.Tensor) -> torch.Tensor: ...




from typing import Iterable, Sequence

import random
import torchdata.datapipes as dp
from functools import partial

from .base import BaseProcessor
from ..fields import SparseField


__all__ = [
    'RandomSource', 'OrderedSource',
    'RandomIDs', 'OrderedIDs', 'DummySource'
]


class RandomSource(BaseProcessor):
    r"""
    DataPipe that generates random items from given source.

    Parameters:
    -----------
    source: Iterable 
        The source data to start.
    datasize: int 
        Datasize.
    """

    def __init__(
        self, source: Iterable, datasize: int
    ) -> None:
        super().__init__(None)

        self._rng = partial(
            random.choice, seq=tuple(source)
        )
        self.datasize = datasize

    def __iter__(self):
        for _ in range(self.datasize):
            yield self._rng()


class OrderedSource(BaseProcessor):
    r"""
    DataPipe that generates ordered items from given source.

    Parameters:
    -----------
    source: Sequence 
        The source data to start.
    datasize: int 
        Datasize.
    """

    def __init__(self, source: Sequence) -> None:
        super().__init__(None)

        assert isinstance(source, Sequence), f"Sequence type is required but received {type(source)} type."
        self.source = tuple(source)

    def __iter__(self):
        yield from iter(self.source)


class RandomIDs(RandomSource):
    r"""
    DataPipe that generates random IDs according to SparseField.

    Parameters:
    -----------
    field: SparseField 
        ID values to select from.
    datasize: int 
        Number of IDs to generate.
    """

    def __init__(
        self, field: SparseField,
        datasize: int,
    ) -> None:
        super().__init__(field.enums, datasize)


class OrderedIDs(OrderedSource):
    r"""
    DataPipe that generates ordered IDs.

    Parameters:
    -----------
    field: SparseField 
        ID values to select from.
    """

    def __init__(self, field: SparseField) -> None:
        super().__init__(field.enums)


class DummySource(OrderedSource):
    r"""
    DataPipe that generates dummy data.

    Parameters:
    -----------
    datasize: int 
        Number of data to generate.
    """
    def __init__(self, datasize: int) -> None:
        super().__init__(range(datasize))


@dp.functional_datapipe("dummy_")
class _DummySource(DummySource):
    r"""
    Functional DataPipe wrapper for DummySource.

    Parameters:
    -----------
    source_dp: dp.iter.IterDataPipe 
        The source DataPipe (No use here).
    datasize: int 
        Number of data to generate.
    """
    def __init__(self, source_dp: dp.iter.IterDataPipe, datasize: int) -> None:
        super().__init__(datasize)




# Grid search for hyper-parameters can be conducted by:
#
#    python -m freerec NAME_OF_EXPERIMENT CONFIG.yaml
#

from .parser import CoreParser
from .launcher import Adapter


cfg = CoreParser()
cfg.compile()

tuner = Adapter()
tuner.compile(cfg)
tuner.fit()

